
import java.util.Observable;

/**
 * Die Klasse Bankkonto stellt die Funktionalitaet eines vereinfachten
 * Bankkontos bereit. Diese beinhaltet das Einzahlen (deposit), das Auszahlen
 * (withdraw) sowie die Berechnung der Zinsen (ohne Beruecksichtigung der
 * Kontobewegungen). Beobachter koennen sich beim Konto anmelden, um Rueckmeldungen
 * ueber Saldoaenderungen zu bekommen.
 */
abstract class Account extends Observable {
	/**
	 * Aktueller Saldo des Bankkontos
	 */
	public float balance;

	/**
	 * Kontonummer des Bankkontos
	 */
	private int accountNo;

	/**
	 * Konstruktor
	 * PRE:  true
	 * POST: true
	 */
	public Account(int accountNo) {
		this.accountNo = accountNo;
	}

	/**
	 * Diese Methode berechnet den aufgelaufenen Zins und fuegt diesen dem
	 * aktuellen Saldo hinzu. Die Berechnung bercksichtigt die 
	 * Kontenbewegungen nicht. Anschliessend werden alle Beobachter dieses Konto 
	 * ueber die Veraenderung des Saldos informiert.
	 * 
	 * PRE   true
	 * POST  balance == balance@PRE + calculateInterests()@PRE 
	 */
	public void accumulateInterests() {
		float oldBalance = balance;
		balance += calculateInterests();
		// Beobachter benachrichtigen
		setChanged();
		// sende den Beobachtern den alten Betrag des Kontos
		notifyObservers(new Float(oldBalance));
	}

	/**
	 * Zahlt auf dem Konto einen gewissen Betrag ein.
	 * Allfaellige Beobachter werden ueber die Saldoaenderung
	 * informiert. Der Betrag darf nicht kleiner sein als null.
	 * 
	 * PRE:  amount >= 0
	 * POST: balance == balance@PRE + amount
	 */
	public void deposit(float amount) {
		float oldBalance = balance;
		balance += amount;
		// Beobachter benachrichtigen
		setChanged();
		notifyObservers(new Float(oldBalance));
	}

	/**
	 * Gibt die Kontonummer zurueck. 
	 * PRE: true 
	 * POST: result == accountNo
	 */
	public int getAccountNo() {
		return accountNo;
	}

	/**
	 * Gibt den den Hashcode dieses Objektes zurueck. Dieser 
	 * ergibt sich aus der (eindeutigen) Kontonummer. 
	 * PRE: true 
	 * POST: result == accountNo
	 */
	public int hashCode() {
		return accountNo;
	}

	/**
	 * Gibt die Informationen (Kontonummer und Saldo) ueber das Konto als String
	 * zurueck. 
	 * PRE: true 
	 * POST: result != null and !result.equals("")
	 */
	public String toString() {
		return "Account " + accountNo + " has balance of " + balance;
	}

	/**
	 * Der angegebene Betrag amount wird vom Saldo abgezogen. Der Betrag darf nicht 
	 * kleiner sein als null. Allfaellige Beobachter des Kontos werden ueber 
	 * die Saldoaenderung informiert. 
	 * 
	 * PRE:  amount >= 0
	 * POST: balance == balance@PRE - amount
	 */
	public abstract void withdraw(float amount);

	/**
	 * Berechnet die aufgelaufenen Zins ohne die Beruecksichtigung der
	 * Kontenbewegungen. 
	 * PRE: true 
	 * POST: true
	 */
	public abstract float calculateInterests();
}