
import java.util.Collections;
import java.util.Enumeration;
import java.util.Observable;
import java.util.Observer;
import java.util.Hashtable;

/**
 * Stellt eine virtuelle Bank dar. 
 * Mit dieser Bank koennen Bankkonten erstellt werden,
 * Betraege eingezahlt, transferiert und abgehoben werden.
 * Es gibt zwei Arten von Bankkonten. Dies sind Sparkonten 
 * und Kontokorrentkonten. Beide unterscheiden sich im 
 * Habenzins und im Betrag den sie ueberzogen werden koennen.
 * Die Bank stellt eine Uebersicht ueber alle Konten und
 * den Gesamtbetrag, der in der Bank eingelagert ist, zur 
 * Verfuegung. 
 */
public class Bank extends Hashtable implements Observer {
	/**
	 * Enthaelt zu jeder Zeit die Summe aller Bankontosalden.
	 * Dieser Gesamtbetrag wird bei Veraenderungen auf einem
	 * Konto nachgefuehrt.
	 */
	private float currentAccountBalance;

	/**
	 * Enthaelt die aktiven Bankkonten.
	 */
	private Hashtable accounts;

	/**
	 * Enthaelt die naechste freie Bankkontonummer.
	 */
	private int nextAccountNo;

	/**
	 * Konstruktor
	 * PRE:  true
	 * POST: true
	 */
	public Bank() {
		currentAccountBalance = 0.0f;
		nextAccountNo = 1;
		accounts = new Hashtable();
	}

	/**
	 * Fuegt ein neues Bankkonto hinzu. 
	 * Das Konto darf noch nicht existieren.
	 */
	void addBankAccount(Account a) {
		currentAccountBalance += a.balance;
		a.addObserver(this);
		accounts.put(new Integer(a.getAccountNo()), a);
	}

	/**
	 * Gibt die Summe der Saldi aller registrierten Konti zurueck.
	 * 
	 * PRE:  true
	 * POST: result >= 0 
	 */
	public float getBankBalance() {
		return currentAccountBalance;
	}
	
	/**
	 * Entfernt ein Bankkonto. Der Saldo der Bank wird dabei nachgefuehrt. 
	 */
	public void removeAccount(Account a) {
		accounts.remove(new Integer(a.getAccountNo()));
		currentAccountBalance -= a.balance;
	}

	/**
	 * Entfernt ein Konto, welches durch die Kontonummer accountNo
	 * identifiziert wird.
	 */
	public void removeAccount(int accountNo) {
		removeAccount(getAccount(accountNo));		
	}
	
	/**
	 * Transferiert den Betrag amount von Konto srcAccount zu Konto
	 * destAccount. 
	 */
	public void transferAmount(int srcAccount, int destAccount, float amount) {
		Account a, b;
		a = getAccount(srcAccount);
		b = getAccount(destAccount);
		a.withdraw(amount);
		b.deposit(amount);
	}

	/**
	 * Zahlt einen Betrag amount auf das Konto mit der Nummer accountNo ein.
	 */
	public void depositMoney(int accountNo, float amount) {
		getAccount(accountNo).deposit(amount);
	}

	/**
	 * Hebt von einem bestimmten Bankkonto mit der Kontonummer accountNo einen
	 * Betrag amount ab.
	 */
	public void withdrawMoney(int accountNo, float amount) {
		getAccount(accountNo).withdraw(amount);
	}

	/**
	 * Gibt ein bestimmtes Bankkonto-Objekt mit der Nummer accountNo zurueck.
	 * PRE: accountExists(accountNo) 
	 * POST: result != null
	 */
	Account getAccount(int accountNo) {
		return (Account) accounts.get(new Integer(accountNo));
	}

	/**
	 * Gibt den Saldo eines bestimmten Kontos aus. 
	 * PRE:  accountExists(accountNo)
	 * POST: result == getAccount(accountNo).balance
	 */
	public float getBalance(int accountNo) {
		return ((Account) accounts.get(new Integer(accountNo))).balance;
	}	

	/**
	 * Gibt true zurueck, falls ein Bankkonto mit der gegebenen Nummer accountNo
	 * existiert. Sonst false. 
	 * 
	 * PRE: true 
	 * POST: if (exists x : Account in
	 *           accounts | x.getAccountNo() == accountNo) 
	 *       then true 
	 *       else false
	 */
	public boolean accountExists(int accountNo) {
		return (accounts.get(new Integer(accountNo)) != null);
	}

	/**
	 * Gibt true zurueck, falls das Bankkonto bereits der Bank zugeordnet ist, ansonsten 
	 * false.
	 * PRE: a != null
	 * POST: if (exists x : Account in accounts  | x == a) then true else false
	 */
	boolean accountExists(Account a) {
		return (accountExists(a.getAccountNo()));
	}
	
	/**
	 * Rueckrufschnittstelle eines Kontobeobachters. Konten, deren Saldo geaendert werden,
	 * melden den alten Saldo ueber diese Methode. Anschliessend wird mittels dem
	 * neuen und dem alten Saldo der neue Gesamtsaldo der Bank berechnet. 
	 * 
	 * PRE:  o != null and arg != null and 
	 *       o instanceof Account and arg instanceof Float 
	 * POST: currentAccountBalance =
	 *         currentAccountBalance@PRE - 
	 *         ((Float)arg).floatValue() +
	 *         ((Account)o).balance
	 */
	public void update(Observable o, Object arg) {
		Float f = (Float) arg;
		currentAccountBalance -= f.floatValue();
		currentAccountBalance += ((Account) o).balance;
	}

	/**
	 * Akkumuliert den aufgelaufenen Zinsbetrag auf allen Konten, ohne
	 * Berueksichtigung der Kontenbewegungen.
	 * PRE:  true
	 * POST: true 
	 */
	public void accumulateInterests() {
		Enumeration e = Collections.enumeration(accounts.values());
		while (e.hasMoreElements()) {
			((Account) e.nextElement()).accumulateInterests();
		}
	}

	/**
	 * Gibt eine Uebersicht ueber alle Konten aus. 
	 * 
	 * PRE: true 
	 * POST: true
	 */
	public void summarizeBank() {
		Enumeration e = Collections.enumeration(accounts.values());
		while (e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("Total amount deposited in bank:" + 
				this.currentAccountBalance);
		System.out.println("*****");
		System.out.println();
	}

	/**
	 * Erzeugt ein neues Kontokorrentkonto und fuegt dieses der Menge von Konten
	 * hinzu. Die Bankkontonummer wird dem Aufrufer zurueckgegeben. 
	 * 
	 * PRE: true
	 * POST: result == nextAccountNo@PRE + 1 and accountExists(result)
	 */
	public int getNewSavingsAccount() {
		SavingsAccount s = newSavingsAccount();
		addBankAccount(s);
		return s.getAccountNo();
	}

	/**
	 * Erzeugt ein neues Sparkonto und fuegt dieses der Menge von Konten hinzu.
	 * Die Bankkontonummer wird dem Aufrufer zurueckgegeben. 
	 * 
	 * PRE: true 
	 * POST: result == nextAccountNo@PRE + 1 and 
	 *       accountExists(result)
	 */
	public int getNewCurrentAccount() {
		CurrentAccount c = newCurrentAccount();
		addBankAccount(c);
		return c.getAccountNo();
	}

	/**
	 * Hilfsmethode zum Erzeugen von neuen Sparkonten.
	 */
	private SavingsAccount newSavingsAccount() {
		SavingsAccount s = new SavingsAccount(nextAccountNo);
		
		nextAccountNo++;
		return s;
	}

	/**
	 * Hilfsmethode zum Erzeugen von Bankkontokorrenten.
	 * PRE:  true
	 * POST: result != null
	 */
	private CurrentAccount newCurrentAccount() {
		CurrentAccount c = new CurrentAccount(nextAccountNo);
		
		nextAccountNo++;
		return c;
	}
}

